// Enable strict semicolon mode
#pragma semicolon 1

// Includes
#include <sourcemod>
#include <sdktools>
#include <tf2>
#include <tf2_stocks>

// Include SMLIB
#include <smlib>

// Constants
#define PLUGIN_VERSION "1.0.1"
#define MAX_WEAPON_LENGTH 128

// Information about the plugin
public Plugin:myinfo =
{
	name = "[TF2] Airshot Bride",
	author = "Jerry",
	description = "Plays a sound upon a successful airshot.",
	version = PLUGIN_VERSION,
	url = "http://jerry.devzero.info/"
}

// CVars
new Handle:CVarSoundFile = INVALID_HANDLE;
new Handle:CVarSoundVolume = INVALID_HANDLE;
new Handle:CVarStopSound = INVALID_HANDLE;
new Handle:CVarHeadshotOnly = INVALID_HANDLE;
new Handle:CVarMinimumHeight = INVALID_HANDLE;

// Weapons that can be used to trigger the airshot sound
new String:AirshotWeapons[][] =
{
	// Rocket Launchers
	"tf_projectile_rocket",             // Rocket Launcher
	"rocketlauncher_directhit",
	"blackbox",
	"liberty_launcher",
	"quake_rl",                         // Original
	"dumpster_device",                  // Beggar's Bazooka
	"cow_mangler",

	// Grenade Launchers
	"tf_projectile_pipe",               // Grenade Launcher
	"loch_n_load",

	// Stickybomb Launchers
	"tf_projectile_pipe_remote",        // Stickybomb Launcher
	"sticky_resistance",                // Scottish Resistance

	// Sniper Rifles
	"sniperrifle",
	"huntsman",
	"sydney_sleeper",
	"bazaar_bargain",
	"machina",
	"pro_rifle",                        // Hitman's Heatmaker

	// Revolvers
	"revolver",
	"samrevolver",                      // Big Kill
	"ambassador",
	"letranger",
	"enforcer",
	"diamondback",

	// Fun stuff
	"market_gardener"
};

// Weapons that must headshot in headshot-only-mode (including weapons that can't headshot)
new String:HeadshotWeapons[][] =
{
	"sniperrifle",
	"huntsman",
	"sydney_sleeper",
	"bazaar_bargain",
	"machina",
	"pro_rifle",                        // Hitman's Heatmaker

	"revolver",
	"samrevolver",                      // Big Kill
	"ambassador",
	"letranger",
	"enforcer",
	"diamondback"
};

// Plugin initialization
public OnPluginStart()
{
	// Hook the player_death event
	HookEvent("player_death", EventPlayerDeath);

	// Create CVars and have SourceMod automatically create the configuration file for us
	CVarSoundFile = CreateConVar("sm_airshot_sound", "misc/airshot_sound.mp3", "Path to the sound file that is played when a player dies to an airshot");
	CVarSoundVolume = CreateConVar("sm_airshot_volume", "1.0", "Volume of the airshot sound (0.0 = completely muted, 1.0 = full volume)", 0, true, 0.0, true, 1.0);
	CVarStopSound = CreateConVar("sm_airshot_stopsound", "1", "If set to 1, the airshot sound will be restarted if it's still playing while another airshot occurs.");
	CVarHeadshotOnly = CreateConVar("sm_airshot_headshot_only", "1", "If set to 1, players using any Sniper Rifle or Revolver *will* have to do headshots for the airshot sound to play.");
	CVarMinimumHeight = CreateConVar("sm_airshot_min_height", "100.0", "The minimum distance above the ground a player needs to be to trigger the airshot sound. Set to 0 to disable.", 0, true, 0.0);
	AutoExecConfig(true, "plugin.airshot-bride");
}

// Sound initialization
public OnConfigsExecuted()
{
	new String:SoundFile[PLATFORM_MAX_PATH];
	new String:FullSoundFile[PLATFORM_MAX_PATH];
	
	// Precache the sound file
	GetConVarString(CVarSoundFile, SoundFile, sizeof(SoundFile));
	PrecacheSound(SoundFile);

	// Prepend "sounds/" and add it to the downloads table
	Format(FullSoundFile, sizeof(FullSoundFile), "sound/%s", SoundFile);
	AddFileToDownloadsTable(FullSoundFile);
}

// Player death (main event)
public Action:EventPlayerDeath(Handle:event, const String:name[], bool:dontBroadcast)
{
	// Attacker and victim
	new attacker = GetClientOfUserId(GetEventInt(event, "attacker"));
	new victim = GetClientOfUserId(GetEventInt(event, "userid"));

	// Weapon stuff
	new String:weapon[MAX_WEAPON_LENGTH];
	new customkill;

	// Sound file
	new String:SoundFile[PLATFORM_MAX_PATH];

	// Height check
	new Float:min_height = GetConVarFloat(CVarMinimumHeight);

	// If the attacker or the victim isn't valid, it might be an environmental death which should be ignored.
	if (!Client_IsValid(attacker) || !Client_IsValid(victim))
		return Plugin_Continue;

	// Ignore people who blow themselves up.
	if (attacker == victim)
		return Plugin_Continue;

	// If they are not in the water and not in the air, they are airborne. And if they're not airborne, it's not an airshot.
	if ((GetEntityFlags(victim) & (FL_INWATER | FL_ONGROUND)))
		return Plugin_Continue;

	// Get the name of the used weapon
	GetEventString(event, "weapon", weapon, sizeof(weapon));

	// Only continue if the weapon can trigger airshots.
	if (!(Array_FindString(AirshotWeapons, sizeof(AirshotWeapons), weapon) >= 0))
		return Plugin_Continue;

	// Check for the required height
	if ((min_height > 0) && (DistanceAboveGround(victim) < min_height))
		return Plugin_Continue;

	// Do we need to check for headshots?
	if (GetConVarBool(CVarHeadshotOnly))
	{
		// Headshot-only-mode is enabled, check whether this weapon can do headshots.
		if (Array_FindString(HeadshotWeapons, sizeof(HeadshotWeapons), weapon) >= 0)
		{
			// It can, or rather, must do headshots to play the airshot sound.
			customkill = GetEventInt(event, "customkill");
			if ((customkill != TF_CUSTOM_HEADSHOT) && (customkill != TF_CUSTOM_HEADSHOT_DECAPITATION) && (customkill != TF_CUSTOM_PENETRATE_HEADSHOT))
			{
				// A headshot is necessary to trigger the sound, but this kill wasn't a headshot, so we return at this point.
				return Plugin_Continue;
			}
		}
	}

	// Get the sound file
	GetConVarString(CVarSoundFile, SoundFile, sizeof(SoundFile));

	// Stop the sound first if necessary
	if (GetConVarBool(CVarStopSound))
		EmitSoundToAll(SoundFile, _, _, _, SND_STOPLOOPING);

	// Play the sound
	EmitSoundToAll(SoundFile, _, _, _, _, GetConVarFloat(CVarSoundVolume));

	return Plugin_Continue;
}

// Calculate a player's distance above the ground.
// Code borrowed from MGE Mod, thanks to Lange!
Float:DistanceAboveGround(client)
{
	decl Float:vStart[3];
	decl Float:vEnd[3];
	new Float:vAngles[3] = {90.0, 0.0, 0.0};
	new Handle:trace;
	new Float:distance = -1.0;

	// Get the client's origin vector and start up the trace ray
	GetClientAbsOrigin(client, vStart);
	trace = TR_TraceRayFilterEx(vStart, vAngles, MASK_SHOT, RayType_Infinite, TraceEntityFilterPlayer);

	if (TR_DidHit(trace))
	{
		// Calculate the distance.
		TR_GetEndPosition(vEnd, trace);
		distance = GetVectorDistance(vStart, vEnd, false);
	}
	else
	{
		// There should always be some ground under the player.
		LogError("[Airshot Bride] Trace error: client %N (%d)", client, client);
	}

	// Clean up and return
	CloseHandle(trace);
	return distance;
}

// Trace ray filter that ignores players
public bool:TraceEntityFilterPlayer(entity, contentsMask)
{
	return !Client_IsValid(entity);
}
